"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import sys

from ._Input import Input


class STDInput(Input):
    """Input handler for sys.stdin"""

    def __init__(self):
        self.closed = True
    
    def name(self):
        return 'stdin'
    
    def open(self):
        self.closed = False
    
    def watch(self, started_callback, output_callback, int_callback):
        self.open()
        started_callback()
        
        while not self.closed:
            try:
                line = sys.stdin.readline()
                if line != '':
                    output_callback(line)
            except:
                pass
            
            self.process_pipe()
    
    def close(self):
        self.closed = True



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division




"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import logging


_config = None


def set_config(cfg):
    global _config
    _config = cfg


def config():
    return _config


class UIPrint(object):
    def __init__(self, notice=None, warning=None, error=None, get_choice=None):
        self.notice = notice or print
        self.warning = warning or print
        self.error = error or print
        self.get_choice = get_choice or print

_ui_print = UIPrint()


def set_ui_print(*args, **kwargs):
    global _ui_print
    _ui_print = UIPrint(*args, **kwargs)


def ui_print():
    return _ui_print


def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)


def get_read_object(args):
    if args.read:
        from snakewatch.input import STD
        return STD.STDInput()
    elif args.watch is not None:
        from snakewatch.input import File
        return File.FileInput(args.watch, args.lines)
    return None


class SysToLogging(object):
    def __init__(self, log_level):
        self.log_level = log_level
        self.linebuf = ''

    def flush(self):
        pass

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            logging.log(self.log_level, line.rstrip())


class AbortError(Exception):
    def __init__(self):
        super(AbortError, self).__init__()


class ConfigError(Exception):
    """Raised when a config error has been detected"""

    def __init__(self, message):
        super(ConfigError, self).__init__(message)


class NotConfirmedError(Exception):
    """Raised when the user has not confirmed a ConfirmAction"""

    def __init__(self):
        super(NotConfirmedError, self).__init__()



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import logging
import os
import re
import six

from colorama import init as cl_init, deinit as cl_deinit, Fore, Back, Style

from .. import config
from ..util import UIPrint, AbortError, get_read_object, ui_print, set_ui_print
from ..main import get_logger, parser

_PREPEND_MSG = '\n{0} *** [{1}{2}{0}] '.format(
    Style.RESET_ALL, Style.DIM,
    'snakewatch.{:d}'.format(os.getpid()),
)

_NOTICE_CLR = Fore.GREEN + Style.DIM
_WARNING_CLR = Fore.YELLOW + Style.BRIGHT
_ERROR_CLR = Fore.RED + Style.BRIGHT

_logger = get_logger('Console')


try:
    _input = raw_input
except NameError:
    _input = input


def setup_arguments(argparser):
    pass


class ConsoleMode(object):
    """UI Handler that outputs to sys.stdout"""

    def __init__(self, *args):
        self.closed = False
        self.cfg = None
        self.received_signal = False
        self.interrupted = False
        self.input = None
        self._waiting_for_input = None
        self.current_style = Style.RESET_ALL
        self.interactive = None
        self.watch_loc = None
        self.config_loc = None

        set_ui_print(self.print_ntc, self.print_warn, self.print_err, self.get_choice)

    def run(self, start_input, args):
        """Read in the config and start watching the input"""
        cl_init()

        if not start_input and not self.watch_loc:
            self.print_err(
                'One of \'--watch\' or \'--read\' must be provided.\n'
                'Use \'--help\' for more information.'
            )
            self.close()
            return

        self.watch_loc = start_input or self.watch_loc

        ui_kwargs = {'ui_confirm': self.confirm}
        if not args.config and not self.config_loc:
            if args.no_config:
                self.cfg = config.DefaultConfig(self, use_file=False, ui_kwargs=ui_kwargs)
            else:
                self.cfg = config.DefaultConfig(self, ui_kwargs=ui_kwargs)
                msg = 'No config provided, using {}'.format(self.cfg.source)
                if self.cfg.source == 'default':
                    msg = '\n'.join([msg, 'Consider creating {}'.format(config.DefaultConfig.file_for(self))])
                self.print_ntc(msg)
        else:
            try:
                self.cfg = config.Config(args.config or self.config_loc, ui_kwargs)
            except AbortError:
                self.close()
                return
            except Exception as err:
                self.print_err('Error in config script {}\n{!s}'.format(
                    args.config, err
                ))
                self.close()
                return

        self.config_loc = self.cfg.source
        
        self.input = start_input
        self.interrupted = False
        self.input.watch(
            self.started_callback, 
            self.output_callback, 
            self.int_callback
        )

        self.close()

    def confirm(self, msg):
        self.print_warn(msg)
        response = 'n'
        prompt = 'Is this OK? [Y/N] : '
        while prompt and not self.received_signal:
            self._waiting_for_input = True
            try:
                response = _input(prompt).lower()
            except EOFError:
                return False
            finally:
                self._waiting_for_input = False

            if response:
                response = response[0]
            prompt = 'Please enter Y or N: ' if response not in ['y', 'n'] else None
        return response == 'y'

    def fatal_error(self, exc_type, exc_value, exc_traceback):
        """Called when an unhandled exception is raised from run()"""
        if not isinstance(exc_value, six.string_types):
            msg = str(exc_value)
        else:
            msg = exc_value

        from snakewatch import LOG_LEVEL

        if LOG_LEVEL == logging.DEBUG and exc_traceback:
            import traceback
            self.print_err(''.join(traceback.format_list(exc_traceback)).rstrip('\n'))

        self.print_err(': '.join([exc_type.__name__, msg]))

    def started_callback(self):
        """Called when the watcher has successfully opened the input"""
        if self.interrupted:
            self.print_ntc('Watch resuming on {}'.format(self.input.name()))
        else:
            self.print_ntc('(press Ctrl-C to stop) Watching {}'.format(self.input.name()))
        self.interrupted = False
        
    def output_callback(self, line):
        """Called when the watcher has read a line and performed an action"""
        self.interrupted = False
        self.cfg.match(line, print, **{'end': ''})

    def int_callback(self, error):
        """Called when the watcher encounters a problem"""
        if not self.interrupted:
            self.print_err('Watch interrupted: {}'.format(error))
        self.interrupted = True
        
    def handle_signal(self, signum, frame):
        """Handle an OS signal from the user"""
        self.received_signal = True

        if self._waiting_for_input:
            self.print_err('Waiting for stdin data. Press enter to quit (use Ctrl/Cmd-D to avoid this message)')
            return

        _logger.debug('Received signal: {:d}'.format(signum))
        if self.input:
            self.input.close()

        self.close()
        
    def close(self):
        """Close any necessary resources"""
        if self.closed:
            return

        self.closed = True
        if self.interactive and not self.input:
            self.interactive.cancel()
        elif self.received_signal:
            self.print_err('Received interrupt, {}'.format('stopping' if self.interactive else 'quitting'))
        
        print(Style.RESET_ALL, end='')
        cl_deinit()

        if self.interactive:
            interactive = self.interactive
            self.__init__()
            self.interactive = interactive

    def get_choice(self, message, prompt, choices, **kwargs):
        self.print_warn(*(message,), **kwargs)
        _choices = []
        choice_range = range(1, len(choices) + 1)
        for i in choice_range:
            _choices.append('{:>2d} - {}'.format(i, choices[i-1]))
        self.print_ntc(*_choices)

        choice = None
        _prompt = prompt
        _colour = _NOTICE_CLR
        while not choice:
            try:
                _choice = _input('{}{}: {}'.format(_colour, _prompt, Style.RESET_ALL))
            except EOFError:
                print()
                return None
            _choice = map(int, re.findall(r'\d+', _choice))[0]
            if _choice not in choice_range:
                _prompt = 'Invalid choice, try again'
                _colour = _ERROR_CLR
                continue
            choice = _choice
        return choices[choice-1]

    def print_msg(self, *args, **kwargs):
        """Print a formatted message"""
        self._print(*args, **kwargs)

    def print_warn(self, *args, **kwargs):
        """Print a formatted warning message"""
        self._print(_WARNING_CLR, *args, **kwargs)

    def print_ntc(self, *args, **kwargs):
        """Print a formatted notice message"""
        self._print(_NOTICE_CLR, *args, **kwargs)
    
    def print_err(self, *args, **kwargs):
        """Print a formatted error message"""
        self._print(_ERROR_CLR, *args, **kwargs)

    def _print(self, colour='', *args, **kwargs):
        try:
            sep = kwargs['sep']
        except KeyError:
            sep = '\n'
        prepend = ''.join([_PREPEND_MSG, colour])
        msg = sep.join(args)
        msg = msg.replace('\n', prepend)
        print(''.join([prepend, msg, self.current_style]))
        print()



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division


def get_arg_group(argparser, mode_name):
    return argparser.add_argument_group('{} Mode Only'.format(mode_name))


def add_cmdln_argument(group, mode_name, *args, **kwargs):
    prefix = mode_name.lower()
    _args = map(list, args)

    for i in range(0, len(args)):
        if args[i][:2] == '--':
            _args[i] = '--{}-{}'.format(prefix, args[i][2:])
        elif args[i][0] == '-':
            _args[i] = '-{}-{}'.format(prefix, args[i][1:])
        else:
            raise ValueError('Mode arguments may not be positional')

    group.add_argument(*_args, **kwargs)



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import re

from colorama import Fore, Back, Style

from ._Action import Action


class PrintAction(Action):
    """An Action that returns the line with possible colouring"""

    non_char = re.compile('[^a-zA-Z]+')
    styleable = ['fore', 'back', 'style']
    
    def __init__(self, cfg):
        super(PrintAction, self).__init__(cfg)
        for part in self.styleable:
            if part in self.cfg:
                self.cfg[part] = self.non_char.sub('', self.cfg[part]).upper()
                
        style = Style.RESET_ALL
        if 'fore' in self.cfg and hasattr(Fore, self.cfg['fore']):
            style = ''.join([style, getattr(Fore, self.cfg['fore'])])
        if 'back' in self.cfg and hasattr(Back, self.cfg['back']):
            style = ''.join([style, getattr(Back, self.cfg['back'])])
        if 'style' in self.cfg and hasattr(Style, self.cfg['style']):
            style = ''.join([style, getattr(Style, self.cfg['style'])])
            
        self.style = style
        
    def run_on(self, line):
        return ''.join([self.style, line])

    def release_resources(self):
        pass



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import importlib
import json
import os
import re
import six

from . import USER_PATH
from .util import AbortError, ConfigError, NotConfirmedError, ui_print, set_config
from .action._ConfirmAction import ConfirmAction


def lower_keys(x):
    """Recursively make all keys lower-case"""
    if isinstance(x, list):
        return [lower_keys(v) for v in x]
    if isinstance(x, dict):
        return dict((k.lower(), lower_keys(v)) for k, v in x.items())
    return x


def find_config(filename):
    if filename.startswith(''.join([os.curdir, os.sep])) or filename.startswith(os.sep):
        return os.path.abspath(filename)
    if filename.startswith('.'):
        return os.path.abspath(os.path.join(USER_PATH, filename[1:]))

    curdircfg = find_config(os.path.join(os.curdir, filename))
    userdircfg = find_config(''.join(['.', filename]))

    if not os.path.exists(curdircfg):
        curdircfg = None
    if not os.path.exists(userdircfg):
        userdircfg = None

    cfg = None
    if curdircfg and userdircfg:
        cfg = ui_print().get_choice(
            'Config filename can refer to multiple locations. Please specify.',
            'Config Location',
            [curdircfg, userdircfg]
        )
    elif curdircfg:
        cfg = curdircfg
    elif userdircfg:
        cfg = userdircfg

    return cfg


class Config(object):
    """An in-memory instance of a user .cfg file"""

    available_actions = {}
    
    def __init__(self, cfg_file, ui_kwargs):
        set_config(self)
        self.actions = []

        if isinstance(cfg_file, six.string_types):
            try:
                fp = open(find_config(cfg_file), 'r')
                self.cfg = json.load(fp)
            except Exception as err:
                ui_print().error('Cannot read config from {}'.format(cfg_file), str(err), sep='\n')
                raise AbortError()

            fp.close()
            self.source = cfg_file
        elif isinstance(cfg_file, list):
            self.cfg = cfg_file
            self.source = 'default'
        self.cfg = lower_keys(self.cfg)
        self.check_actions(ui_kwargs)
        
    def check_actions(self, ui_kwargs):
        """Create an action instance for each entry in the config"""
        ptrn = re.compile('[\W]+')

        for entry in self.cfg:
            name = ptrn.sub('', entry['action']).title()
            entry['action'] = name

            action_object_name = '{}Action'.format(name)

            try:
                action = Config.available_actions[name]
            except KeyError:
                action_module = importlib.import_module('snakewatch.action.{}'.format(name))
                Config.available_actions[name] = action = getattr(action_module, action_object_name)

            kwargs = {
                'cfg': entry,
            }
            if issubclass(action, ConfirmAction):
                kwargs['ui_confirm'] = ui_kwargs['ui_confirm']

            try:
                self.actions.append(action(**kwargs))
            except ConfigError as ce:
                ui_print().error('Config error detected:', str(ce), sep='\n')
                raise AbortError()
            except NotConfirmedError as nce:
                ui_print().error('Config entry not confirmed')
                raise AbortError()
            except:
                ui_print().error('Fatal Error when instantiating {}'.format(action_object_name))
                raise
        
    def match(self, line, output_method, **output_kwargs):
        """Perform the first action where action.matches(line) is True

        If a match is found, return the result of action.run_on(line)
          or '' if the result is None

        If no match is found, return the line unchanged
        """
        matched = False
        for action in self.actions:
            if not action.matches(line):
                continue

            matched = True
            result = action.run_on(line)
            if result is not None:
                output_method(result, **output_kwargs)

            if not action.continue_matching():
                break

        if not matched:
            output_method(line, **output_kwargs)


class DefaultConfig(Config):
    """A default default config.

    This config is only used when <USER_PATH>/default.json does not exist,
    and uses Print for all inputs.
    """
    def __init__(self, ui, ui_kwargs, use_file=True):
        user_default = DefaultConfig.file_for(ui)
        if use_file and os.path.exists(user_default):
            super(DefaultConfig, self).__init__(user_default, ui_kwargs)

        cfg = [
            {
                'regex': '.*',
                'action': 'Print',
            },
        ]
        super(DefaultConfig, self).__init__(cfg, ui_kwargs)

    @classmethod
    def file_for(cls, ui):
        return os.path.join(USER_PATH, 'default-{}.json'.format(ui.__class__.__name__))



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import os

from ._ConfirmAction import ConfirmAction
from ..util import AbortError, ConfigError, ui_print


class WriteAction(ConfirmAction):
    """An Action that returns the line with possible colouring"""

    instances = dict()

    def __init__(self, cfg, ui_confirm):
        self.mode = 'w' if cfg.get('truncate', False) else 'a'

        if 'filename' in cfg:
            filename = cfg['filename']
            if filename.startswith('~'):
                filename = os.path.expanduser(filename)
            self.filename = os.path.abspath(filename)

        super(WriteAction, self).__init__(cfg, ui_confirm, ['filename'])

        WriteAction.open_file_instance(self)

    @classmethod
    def open_file_instance(cls, inst):
        try:
            file_instances = cls.instances[inst.filename]
        except KeyError:
            file_instances = list()
            cls.instances[inst.filename] = file_instances

        if file_instances:
            inst.fp = file_instances[0]

            if inst.fp.mode != inst.mode:
                raise ConfigError('File {} is opened in conflicting modes.'.format(inst.filename))
        else:
            try:
                inst.fp = open(inst.filename, inst.mode)
            except (OSError, IOError) as err:
                ui_print().error(
                    'Cannot open {} for writing.'.format(inst.filename),
                    str(err), sep='\n'
                )
                raise AbortError()

        file_instances.append(inst)

    @classmethod
    def close_file_instance(cls, inst):
        try:
            file_instances = cls.instances[inst.filename]
        except KeyError:
            return

        try:
            file_instances.remove(inst)
        except ValueError:
            pass

        if not file_instances:
            inst.fp.close()

    def run_on(self, line):
        self.fp.write(line)
        self.fp.flush()
        os.fsync(self.fp)
        return None

    def release_resources(self):
        WriteAction.close_file_instance(self)

    def confirm_message(self):
        return 'The file {} will be {}.'.format(
            self.filename,
            'overwritten' if self.mode == 'w' else 'written to',
        )



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import argparse
import importlib
import logging
import os
import signal
import stat
import sys

from logging.handlers import RotatingFileHandler

from . import (
    NAME, VERSION, DESCRIPTION, USER_PATH, URL, AUTHOR, AUTHOR_EMAIL,
    LOG_FILE, LOG_LEVEL, LOG_BACKUP_COUNT, LOG_MAX_BYTES, LOG_FORMAT, LOG_TO_STDOUT
)
from .util import AbortError, get_read_object, config, ui_print


_logger = logging.getLogger()
_logger.setLevel(LOG_LEVEL)

_log_handler = logging.StreamHandler()
_log_handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT))
_logger.addHandler(_log_handler)

parser = None


def get_logger(name):
    """Get a logging instance consistent with the main logger"""
    if sys.version_info > (2, 7):
        return _logger.getChild(name)
    return logging.getLogger('.'.join([_logger.name, name]))


def release_action_resources():
    """Release all resources loaded by all actions"""
    if config() is None:
        return

    for action in config().actions:
        try:
            action.release_resources()
        except:
            ui_print().error(
                'Unable to release resources for action {}'.format(action.__class__.__name__),
                str(action.cfg), sep='\n'
            )


def main(initial_args=None, handle_signals=True):
    global _log_handler, parser

    if initial_args is None:
        initial_args = sys.argv[1:]

    log_to_file = True
    if not os.path.exists(USER_PATH):
        try:
            os.makedirs(USER_PATH)
        except OSError:
            log_to_file = False
            print('Unable to create snakewatch settings/log directory.',
                  'Please create the directory {}'.format(USER_PATH),
                  sep='\n', file=sys.stderr)

    if not os.access(USER_PATH, os.W_OK):
        try:
            mode = stat.S_IWRITE
            if not sys.platform == 'win':
                st = os.stat(USER_PATH)
                mode = mode | st.mode
            os.chmod(USER_PATH, mode)
        except OSError:
            log_to_file = False
            print('Unable to write to snakewatch settings/log directory.',
                  'Please set write permissions to the directory {}'.format(USER_PATH),
                  sep='\n', file=sys.stderr)

    if log_to_file and not LOG_TO_STDOUT:
        _logger.removeHandler(_log_handler)
        _log_handler.close()

        _log_handler = RotatingFileHandler(
            filename=LOG_FILE,
            maxBytes=LOG_MAX_BYTES,
            backupCount=LOG_BACKUP_COUNT,
        )

        _log_handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT))
        _logger.addHandler(_log_handler)
    
    parser = argparse.ArgumentParser(
        prog=NAME,
        description=DESCRIPTION,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    global_args = parser.add_argument_group('Global')
    parser.add_argument(
        '-v', '--version', 
        action='version', 
        version='\n'.join([NAME, VERSION, '', '{} <{}>'.format(AUTHOR, AUTHOR_EMAIL), URL])
    )
    parser_config = global_args.add_mutually_exclusive_group()
    parser_config.add_argument(
        '-c', '--config', 
        help='which configuration file to use'
    )
    parser_config.add_argument(
        '--no-config', action='store_true',
        help='don\'t use any configuration file (including the default), print everything'
    )
    parser.add_argument(
        '-n', '--lines',
        default=0, type=int,
        help='start LINES from end of the file, use -1 to start at the beginning',
    )

    watch_loc_group = global_args.add_mutually_exclusive_group()
    watch_loc_group.add_argument(
        '-w', '--watch', 
        help='which file to watch'
    )
    watch_loc_group.add_argument(
        '-r', '--read',
        action='store_true',
        help='read input from stdin'
    )

    # Only one mode for now, so exclude all this stuff.

    # import imp
    # import importlib
    # from snakewatch import mode as mode_package
    # suffixes = tuple([suffix[0] for suffix in imp.get_suffixes()])
    # mode_names = set([
    #     os.path.splitext(module)[0]
    #     for module in os.listdir(mode_package.__path__[0])
    #     if module.endswith(suffixes)
    # ])
    # available_modes = dict()
    # for mode_name in mode_names:
    #     if mode_name == '__init__':
    #         continue
    #     try:
    #         mode_module = importlib.import_module('snakewatch.mode.{}'.format(mode_name))
    #     except ImportError:
    #         _logger.exception('Could not load mode module {}'.format(mode_name))
    #         continue
    #     else:
    #         available_modes[mode_name] = mode_module
    #
    #     setup_arguments = getattr(mode_module, 'setup_arguments', None)
    #     if setup_arguments and callable(setup_arguments):
    #         try:
    #             setup_arguments(parser)
    #         except:
    #             _logger.exception('{} mode has arguments but setup failed'.format(mode_name))
    #
    # if not available_modes:
    #     _logger.critical('No modes are available')
    #     return 1

    # parser.add_argument(
    #     '-m', '--mode',
    #     choices=available_modes,
    #     default='Console',
    #     help='which mode to use'
    # )

    try:
        args = parser.parse_args(initial_args)
    except SystemExit:
        return

    args.mode = 'Console'

    _logger.debug('{}\n'.format('=' * 40))

    mode = importlib.import_module('.mode.Console', __package__)
    handler = getattr(mode, '{}Mode'.format(args.mode), None)
    if not handler or not callable(handler):
        _logger.critical('{} mode structure is not valid'.format(args.mode))
        sys.exit(1)

    handler = handler()

    if handle_signals:
        if not sys.platform.startswith('win'):
            signal.signal(signal.SIGHUP, handler.handle_signal)
            signal.signal(signal.SIGQUIT, handler.handle_signal)
        signal.signal(signal.SIGINT, handler.handle_signal)
        signal.signal(signal.SIGTERM, handler.handle_signal)
        signal.signal(signal.SIGABRT, handler.handle_signal)

    try:
        handler.run(get_read_object(args), args)
    except AbortError:
        pass
    except:
        if LOG_LEVEL == logging.DEBUG:
            raise
        import traceback
        exc_type, exc_value = sys.exc_info()[:2]
        exc_traceback = traceback.extract_stack()
        handler.fatal_error(exc_type, exc_value, exc_traceback)
        return 1
    finally:
        release_action_resources()

    _logger.debug('snakewatch exiting\n')
    _log_handler.close()

if __name__ == '__main__':
    sys.exit(main() or 0)



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import copy
import re
import six

from abc import ABCMeta, abstractmethod

from ..util import ConfigError


@six.add_metaclass(ABCMeta)
class Action(object):
    """Base class for all Actions"""

    def __init__(self, cfg, required_attributes=list()):
        self.raw_cfg = copy.deepcopy(cfg)
        self.cfg = cfg
        try:
            self.pattern = re.compile(self.cfg['regex'])
        except Exception as err:
            self._config_error('Invalid RegEx: {!s}'.format(err))

        self.name = self.__module__.split('.')[-1:][0]
        self.stop_matching = 'continue' not in self.cfg or not self.cfg['continue']

        for attr in required_attributes:
            if attr not in cfg:
                self._config_error('missing required attribute {}'.format(attr))

    def _config_error(self, message):
        raise ConfigError('{}: {}\n{}'.format(
            self.__class__.__name__, message, self.cfg
        ))

    def matches(self, line):
        return self.pattern.match(line) is not None

    def continue_matching(self):
        return not self.stop_matching

    @abstractmethod
    def run_on(self, line):
        pass

    @abstractmethod
    def release_resources(self):
        pass



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import six
import os
import time

from ._Input import Input


class FileInput(Input):
    """An Input that reads from a system file"""

    def __init__(self, filename, readback=0):
        self.filename = filename
        self.readback = readback
        self.reopen = True
        self.has_opened = False
        self.fp = None
        self.where = 0
    
    def name(self):
        return os.path.basename(self.filename)
    
    def open(self):
        self.fp = open(self.filename, 'r')
        if not self.has_opened:
            self.has_opened = True
            if self.readback > -1:
                self.fp.seek(0, os.SEEK_END)
                if self.readback > 0:
                    hit_start = False
                    while self.readback > 0:
                        if self.fp.read(1) == '\n':
                            self.readback -= 1
                        try:
                            self.fp.seek(-2, os.SEEK_CUR)
                        except IOError:
                            hit_start = True
                            break

                    if hit_start:
                        self.fp.seek(0)
                    else:
                        self.fp.seek(2, os.SEEK_CUR)

    def watch(self, started_callback, output_callback, int_callback, poll_callback=None):
        while self.reopen:
            try:
                if poll_callback:
                    poll_callback()
                self.open()
            except Exception as err:
                int_callback('\n'.join([self.filename, err]))
                time.sleep(1)
                self.fp = None
            else:
                started_callback()
            
            while self.fp and not getattr(self.fp, 'closed', True):
                if poll_callback:
                    poll_callback()

                line = self.readline(int_callback)
                if line != '':
                    output_callback(line)
                else:
                    time.sleep(0.1)
    
    def readline(self, int_callback):
        """Read a line from the file.

        If the file has been truncated, or cannot be read from, close the file handle,
        set the readback to start at the beginning, and let the watcher re-open the file.
        """
        try:
            fs = os.stat(self.filename)
            self.where = self.fp.tell()
            if fs.st_size < self.where:
                # File contents has been truncated, so close and reopen the file
                self.re_open()
                return ''
            line = self.fp.readline()
        except Exception as err:
            int_callback('\n'.join([self.filename, err]))
            self.re_open()
            return ''
        else:
            return line

    def re_open(self):
        self.close()
        self.readback = -1
        self.has_opened = False
        self.reopen = True

    def close(self):
        self.reopen = False
        if not self.fp:
            return
        if six.callable(getattr(self.fp, 'close', None)):
            self.fp.close()
        self.fp = None



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import os
import logging

NAME = 'snakewatch'
VERSION = '1.0.0'
DESCRIPTION = '{} v{}\nA log watcher'.format(NAME, VERSION)
URL = 'https://github.com/asoc/snakewatch'
AUTHOR = 'Alex Honeywell'
AUTHOR_EMAIL = 'alex.honeywell@gmail.com'

USER_PATH = os.path.expanduser(os.path.join('~', '.snakewatch'))

LOG_FILE = os.path.join(USER_PATH, 'snakewatch.log')
LOG_TO_STDOUT = False
LOG_FORMAT = '%(asctime)-15s [%(levelname)s] %(name)s: %(message)s'
LOG_LEVEL = logging.INFO
LOG_MAX_BYTES = 1024*1024*5
LOG_BACKUP_COUNT = 1



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function, absolute_import, unicode_literals, division

import os

from setuptools import setup, find_packages

from snakewatch import NAME, VERSION, DESCRIPTION, URL, AUTHOR, AUTHOR_EMAIL

extra_options = dict(
)
options = dict(
)


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as fp:
        return fp.read()


setup(
    name=NAME,
    version=VERSION,
    
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    
    url=URL,
    description=DESCRIPTION,
    keywords='log, tail',
    long_description=read('README'),

    packages=find_packages(),
    install_requires=['colorama', 'six'],
    
    entry_points={
        'console_scripts': [
            'snakewatch = snakewatch.main:main',
        ],
    },

    license='LGPL',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: System :: Logging',
    ],

    options=options,
    **extra_options
)



"""
This file is part of snakewatch.

snakewatch is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

snakewatch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with snakewatch.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import six

from abc import ABCMeta, abstractmethod, abstractproperty


@six.add_metaclass(ABCMeta)
class Input(object):
    """Base class for all Inputs"""

    pipe = None

    def connect_process(self, pipe):
        """Connect this input to the receiving end of a Pipe"""
        self.pipe = pipe
        
    def process_pipe(self):
        """Process any signals from the Pipe"""
        if self.pipe is not None:
            signal = None
            try:
                if self.pipe.poll():
                    signal = self.pipe.recv()
            except Exception:
                pass
            if signal == 'close':
                self.close()
    
    @abstractproperty
    def name(self):
        """Name of the Input"""
        return None
    
    @abstractmethod
    def open(self):
        """Open the Input and get ready for watching"""
        pass
    
    @abstractmethod
    def watch(self, started_callback, output_callback, int_callback):
        """Start watching the Input"""
        pass
    
    @abstractmethod
    def close(self):
        """Close the Input and any associated resources"""
        pass
